
/*
  
  File audio.c - routines for 2352 byte frames cdfs

  
  Copyright (c) 1999,2000 by Michiel Ronsse 
  
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
  
*/


#include "cdfs.h"


/***************************************************************************/

void make_header(char * temp, unsigned int size){
    PRINT("MAKING HEADER\n");

    /* RIFF header */
    strcpy(             temp,     "RIFF");
    *(unsigned int*)   (temp+ 4) = size-8;    /* size of RIFF */

    /* WAVE file */
    strcpy(             temp+ 8,  "WAVE");

    /* WAVE format*/
    strcpy(             temp+12,  "fmt ");
    *(unsigned int*)   (temp+16) = 16;        /* size of format descriptor */
    /* common fields */
    *(unsigned short*) (temp+20) = 1;         /* format category: WAVE_FORMAT_PCM */
    *(unsigned short*) (temp+22) = 2;         /* number of channels */
    *(unsigned int*)   (temp+24) = 44100;     /* sample rate */
    *(unsigned int*)   (temp+28) = 44100*4;   /* bytes/sec */
    *(unsigned short*) (temp+32) = 4;         /* data block size */
    /* format specific fields (PCM) */
    *(unsigned short*) (temp+34) = 16;        /* bits per sample */

    /* WAVE data */
    strcpy(             temp+36,  "data");
    *(unsigned int*)   (temp+40) = size-44;   /* size of data chunk */

}
/***************************************************************************/

void copy_from_cd(struct super_block * sb, int inode, unsigned int start, 
                  unsigned int stop, char * buf){
  int start_sector, start_byte, stop_sector, stop_byte, sector;
  int status;
  struct cdrom_read_audio cdda;
  unsigned int read_size=CD_FRAMESIZE_RAW;
  cd * this_cd = sb->u.generic_sbp;      
  unsigned start_lba=this_cd->track[inode-3].start_lba;
  unsigned stop_lba = this_cd->track[inode-3].stop_lba;

  /* cache */
  char * temp;
  char * temp_ptr;
  char * temp_start;
  int temp_length;
  char * buf_ptr;

  if (!this_cd->audio_cache) {    /* nog geen cache: eerste gebruik */
    this_cd->sector_cache=-READ_AHEAD-1;     /* empty cache */
    if (!(this_cd->audio_cache = kmalloc((1+READ_AHEAD)*CD_FRAMESIZE_RAW, GFP_KERNEL))){
      printk("Could not allocate cache\n");
      return;    
    }
  }
  
  temp=this_cd->audio_cache;

  start_sector = start/read_size;
  start_byte   = start - start_sector*read_size;
  stop_sector  = stop/read_size;
  stop_byte    = stop - stop_sector*read_size;
  
  start_sector += start_lba;
  stop_sector  += start_lba;
  
  if (!stop_byte) {            /* empty frame */
    stop_sector -= 1;
    stop_byte    = CD_FRAMESIZE_RAW;
  }
  
  PRINT("%d[%d-%d] -> 0x%x...0x%x  ... (%d,%d),(%d,%d)\n", 
        inode, start, stop, (int)buf, (int)buf+stop-start,
        start_sector,start_byte,stop_sector,stop_byte);
  
  buf_ptr=buf;
  
  for (sector=start_sector; sector<=stop_sector; sector++){
    
    PRINT("cache=[%d-%d] reading sector=%d\n", this_cd->sector_cache, this_cd->sector_cache+READ_AHEAD, sector);
    
    if (!((this_cd->sector_cache<=sector) && (sector<=this_cd->sector_cache+READ_AHEAD))) {
      
      PRINT("reading sectors %d..%d\n", sector, sector+READ_AHEAD);
      cdda.addr.lba    = sector;
      cdda.addr_format = CDROM_LBA;
      cdda.nframes     = 1+READ_AHEAD;
      cdda.buf         = temp;
      this_cd->sector_cache     = sector;
      temp_ptr         = temp;
      
      if (cdda.nframes > stop_lba-sector)     /* reduce cache if necessary */
        cdda.nframes=stop_lba-sector;
     
      status = cdfs_ioctl(sb, CDROMREADAUDIO, (unsigned long)&cdda);
      if (status) {
        printk("ioctl failed: %d, address=%d\n", status, cdda.addr.lba);
        return;
      }
      
    } else {
      PRINT("getting sector %d from cache\n", sector);
      temp_ptr = temp + (sector-this_cd->sector_cache)*CD_FRAMESIZE_RAW;
    }
    
    if (sector==start_sector) {
      temp_start  = temp_ptr+start_byte;
      if (sector!=stop_sector) 
        temp_length = read_size-start_byte;
      else
        temp_length = stop_byte-start_byte;
    } else if (sector==stop_sector) {
      temp_start  = temp_ptr;
      temp_length = stop_byte;
    } else {
      temp_start  = temp_ptr;
      temp_length = read_size;
    }
    
    PRINT("memcpy(0x%x, %x, %d)\n",(int)buf_ptr, (int)temp_start, temp_length);
    copy_to_user(buf_ptr, (char*)temp_start, temp_length);
    buf_ptr+=temp_length;
    
  }
};

/***************************************************************************/

static ssize_t cdda_file_read(struct file * file, char * buf, size_t count, loff_t *ppos){
  int size;
  struct dentry * dentry = file->f_dentry;
  struct inode * inode = dentry->d_inode;  
  unsigned start,stop;

  PRINT("cdda_file_read(file=0x%x, buf=0x%x, count=%d, *ppos=%d)\n",
         (int)file, (int)buf, count, *(int*)ppos);


  if (file->f_pos < inode->i_size) {
    if (count > inode->i_size-file->f_pos) 
      size = inode->i_size-file->f_pos;
    else
      size = count;
  } else {
    return 0;
  }
  
  /* size > 0 */

  start = file->f_pos;
  stop  = start+size;

  PRINT("--> file->f_pos=%d, inode->i_size=%d, size=%d\n", 
        (int)file->f_pos, (int)inode->i_size, size);
  
  if (start < WAV_HEADER_SIZE) {    /* we lezen (een deel van) de header ! */
    if (stop > WAV_HEADER_SIZE) {   /* we lezen ook een deel van de data => splitsen */
      cdda_file_read( file,  buf,                        WAV_HEADER_SIZE-start,  ppos );
      cdda_file_read( file,  buf+WAV_HEADER_SIZE-start,  stop-WAV_HEADER_SIZE,   ppos );
      return size;
    } else {                        /* enkel header */
      char temp[44];
      make_header(temp, inode->i_size);
      copy_to_user(buf, temp+start, size);
    }
  } else {                          /* enkel data */
    start -= WAV_HEADER_SIZE;
    stop  -= WAV_HEADER_SIZE;
    copy_from_cd(inode->i_sb, inode->i_ino, start, stop, buf);
  }

  file->f_pos += size;
  return size;
}

/***************************************************************************/

struct file_operations cdda_file_operations = {
  NULL,			/* lseek */
  cdda_file_read,	/* read */
  NULL,			/* write */
  NULL,			/* readdir */
  NULL,			/* poll */
  NULL,			/* ioctl */
  NULL,			/* mmap */
  NULL,			/* open */
  NULL,			/* flush */
  NULL,			/* release */
  NULL,			/* fsync */
  NULL,			/* fasync */
  NULL,			/* check_media_change */
  NULL			/* revalidate */
};

struct inode_operations cdda_file_inode_operations = {
  &cdda_file_operations,
  NULL,			/* create */
  NULL,			/* lookup */
  NULL,			/* link */
  NULL,			/* unlink */
  NULL,			/* symlink */
  NULL,			/* mkdir */
  NULL,			/* rmdir */
  NULL,			/* mknod */
  NULL,			/* rename */
  NULL,			/* readlink */
  NULL,			/* follow_link */
  NULL,
  NULL,			/* writepage */
  NULL,			/* bmap */
  NULL,			/* truncate */
  NULL,			/* permission */
  NULL,			/* smap */
};

