/*
  
  File cdXA.c - routines for frames <>2048 and <> 2352 byte frames for cdfs

  
  Copyright (c) 1999,2000 by Michiel Ronsse 
  
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
  
*/


#include "cdfs.h"



/*
 * format            sector type                 user data size (bytes)            ioctl
 * -----------------------------------------------------------------------------------------
 *   1     (Red Book / audio)    CD-DA          2352    (CD_FRAMESIZE_RAW)         CDROMREADAUDIO / CDROMREADRAW
 *   2     (Yellow Book / data)  Mode1 Form1    2048    (CD_FRAMESIZE)             CDROMREADMODE1
 *   3     (Yellow Book / data)  Mode1 Form2    2336    (CD_FRAMESIZE_RAW0)        CDROMREADMODE2
 *   4     (Green Book / XA)     Mode2 Form1    2048    (CD_FRAMESIZE)
 *   5     (Green Book / XA)     Mode2 Form2    2328    (2324+4 spare bytes)           
 *                                              2646    (CD_FRAMESIZ_RAWER)
 *                                              2340    (CD_FRAMESIZE_RAW1)
 */

struct _tracktype_ {
  char name[32];
  unsigned int framesize;
  unsigned int ioctl;
} type[5] = {
  {"Red Book / Audio / CD-DA", CD_FRAMESIZE_RAW, CDROMREADAUDIO},
  {"Yellow Book / Data / Mode1 Form1",CD_FRAMESIZE , CDROMREADMODE1},
  {"Yellow Book / Data / Mode1 Form2",CD_FRAMESIZE_RAW0 , CDROMREADMODE2},
  {"Green Book / Data / Mode2 Form1",CD_FRAMESIZE , 0},
  {"Green Book / Data / Mode2 Form2",2324 , 0}
};

static ssize_t cdXA_file_read(struct file * file, char * buf, size_t count, loff_t *ppos);

struct file_operations cdXA_file_operations = {
  NULL,
  cdXA_file_read,
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL
};

struct inode_operations cdXA_file_inode_operations = {
  &cdXA_file_operations,
  NULL,			/* create */
  NULL,			/* lookup */
  NULL,			/* link */
  NULL,			/* unlink */
  NULL,			/* symlink */
  NULL,			/* mkdir */
  NULL,			/* rmdir */
  NULL,			/* mknod */
  NULL,			/* rename */
  NULL,			/* readlink */
  NULL,			/* follow_link */
  NULL,
  NULL,			/* writepage */
  NULL,		/* bmap */
  NULL,			/* truncate */
  NULL,			/* permission */
  NULL,			/* smap */
};

/***********************************************************/


void copy_from_cdXA(struct super_block * sb, int inode, unsigned int start, 
                  unsigned int stop, char * buf, unsigned int track_size){
  int start_sector, start_byte, stop_sector, stop_byte, sector;
  int status;
  unsigned int read_size=CD_FRAMESIZE_RAW0;
  cd * this_cd = sb->u.generic_sbp;      
  unsigned start_lba=this_cd->track[inode-3].start_lba;

  /*static*/ char temp[CD_FRAMESIZE_RAW0];
  /*static*/ int last_sector=-1;


  struct cdrom_msf  * msf = (struct cdrom_msf*)temp;

  char * temp_start;
  int temp_length;

  char * buf_ptr;

  PRINT("start_lba=%d\n", start_lba);

  start_sector = start/read_size;
  start_byte   = start - start_sector*read_size;
  stop_sector  = stop/read_size;
  stop_byte    = stop - stop_sector*read_size;

  PRINT("%d[%d-%d] -> 0x%x...0x%x  ... (%d,%d),(%d,%d)\n", 
         inode, start, stop, (int)buf, (int)buf+stop-start,
         start_sector,start_byte,stop_sector,stop_byte);
 
  
  buf_ptr=buf;
  
  for (sector=start_sector; sector<=stop_sector; sector++){
    
      int lba=sector+start_lba;

      PRINT("reading sector %d, lba=%d\n", sector, lba);

      if (last_sector == lba) {
        PRINT("using cache\n");
      } else {
        last_sector=lba;

        msf->cdmsf_frame0=lba % 75;
        lba /= 75;
        /*        lba += 2;*/
        msf->cdmsf_sec0=lba%60;
        msf->cdmsf_min0=lba/60;
     
        PRINT("%d:%d:%d\n", msf->cdmsf_min0, msf->cdmsf_sec0, msf->cdmsf_frame0);

        status = cdfs_ioctl(sb, CDROMREADMODE2, (unsigned long)msf);
        if (status) {
          printk("ioctl failed: %d\n", status);
          return;
        }
     }
    
    if (sector==start_sector) {
      temp_start  = temp+start_byte;
      if (sector!=stop_sector) 
        temp_length = read_size-start_byte;
      else
        temp_length = stop_byte-start_byte;
    } else if (sector==stop_sector) {
      temp_start  = temp;
      temp_length = stop_byte;
    } else {
      temp_start  = temp;
      temp_length = read_size;
    }
    
    PRINT("memcpy(0x%x, %x, %d)\n", (int)buf_ptr, (int)temp_start, temp_length);
    
    copy_to_user(buf_ptr, (char*)temp_start, temp_length);
    
    buf_ptr+=temp_length;

  }
};

/***************************************************************************/

static ssize_t cdXA_file_read(struct file * file, char * buf, size_t count, loff_t *ppos){
  int size;
  struct dentry * dentry = file->f_dentry;
  struct inode * inode = dentry->d_inode;  
  
  PRINT("cdXA_file_read(file=0x%x, buf=0x%x, count=%d, *ppos=%d)\n",
         (int)file, (int)buf, count, *(int*)ppos);


  if (file->f_pos < inode->i_size ) {
    if (count > inode->i_size-file->f_pos) 
      size = inode->i_size-file->f_pos;
    else
      size = count;
  } else {
    return 0;
  }
  
  /* size > 0 */

  PRINT("--> file->f_pos=%d, inode->i_size=%d, size=%d\n", 
         (int)file->f_pos, (int)inode->i_size, size);

  copy_from_cdXA(inode->i_sb, inode->i_ino, file->f_pos, file->f_pos+size, buf, inode->i_size);

  file->f_pos+=size;
  return size;
}
